/* ReSCAL - Main entry
 *
 * Copyright (C) 2011-2012
 *
 * Author: Olivier Rozier <rozier@ipgp.fr>
 *
 * This file is part of ReSCAL.
 *
 * ReSCAL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define _MAIN_
#include <locale.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "defs.h"
#include "macros.h"
#include "entry.h"
#include "param.h"
#include "cells.h"
#include "simul.h"
#include "interface.h"
#include "callbacks.h"
//#include "monitor.h"
#include "view.h"
#include "rescal.h"
#include "space.h"
#include "templates.h"
#include "format.h"
#include "schedul.h"
#ifdef PARALLEL
#include "messages.h"
#endif


extern int graine;            // graine pour la generation des nombres aleatoires
extern float dump_delay_png;
extern float dump_delay_csp;
extern unsigned char csphpp_flag;
extern int flag_save_auto;
extern int flag_dcsp_auto;
extern int flag_dhpp_auto;
extern int flag_dpng_auto;
extern double stop_delay_t0;
extern double stop_time;
extern int mode_pat;
#ifdef PARALLEL
extern int mode_par;    //mode parallele
extern int proc_id;
#endif

int prog=PROG_RESCAL;
int arch=0;
char* comp_date=__DATE__;
float frame_delay=0;
unsigned char opt_h=0, opt_hm=0, opt_ht=0, opt_nv=0, opt_log=0,  opt_dcsp=0, opt_dhpp=0, opt_dbin=0, opt_stop=0, opt_quit=0;
DumpDelay dcsp_delay, dhpp_delay, dpng_delay, dlog_delay;
unsigned char flag_no_input=0;
gint dbin_delay, stop_delay;
unsigned char opt_dpng=0, opt_djpeg=0;
gint djpeg_delay=0;

int main (int argc, char *argv[])
{
  DumpPar dp_info, dp_csp, dp_bin, dp_png, dp_jpeg;

#ifdef PARALLEL
  if (argc>1) msg_init();
#endif

#ifdef LOG_FILE
  char log_filename[100];
  sprintf(log_filename,"RESCAL.log");
#ifdef PARALLEL
  if (proc_id>=0){
    sprintf(log_filename,"RESCAL%d.log",proc_id);
  }
#endif
  if (argc>1) log_file = fopen(log_filename,"w");
#endif

  //LogPrintf("ReSCAL\n");
  LogPrintf("ReSCAL %s (%s)\n",VER_NUM,VER_DAT);
#ifdef VER_BRA
  LogPrintf("branch %s\n",VER_BRA);
#endif
  LogPrintf("%s model", MOD_NAME);
#ifdef MOD_DESC
  LogPrintf(" - %s", MOD_DESC);
#endif
  LogPrintf("\n");
  LogPrintf("compiled: %s (", comp_date);
#ifdef __INTEL_COMPILER
  LogPrintf("icc %d", __INTEL_COMPILER);
#elif defined(__clang__)
  LogPrintf("clang %d.%d.%d", __clang_major__, __clang_minor__, __clang_patchlevel__);
#elif defined(__GNUC__)
  LogPrintf("gcc %d.%d", __GNUC__, __GNUC_MINOR__);
#endif
#ifdef GTK2
  LogPrintf(", gtk 2.%d", GTK_MINOR_VERSION);
#endif
#ifdef USE_GD
  LogPrintf(", gd %d.%d", GD_MAJOR_VERSION, GD_MINOR_VERSION);
#endif
  LogPrintf(")\n");

  gtk_set_locale();
#ifdef NUM_MODE
  LogPrintf("format of numerical values: %s \n", NUM_MODE);
  setlocale(LC_NUMERIC, NUM_MODE);
#endif
  if (sizeof(long)==4) arch = W32;
  if (sizeof(long)==8) arch = W64;
  LogPrintf("size of long: %ld bytes (%s)\n", sizeof(long), (arch==W32)?"32 bits":"64 bits");
  //LogPrintf("size of int: %ld bytes (%s)\n", sizeof(int), (sizeof(int)==4)?"32b":"64b");
  LogPrintf("size of cell: %lu byte(s)\n", sizeof(Cell));
#ifndef GUI
  LogPrintf("graphical interface disabled\n");
#endif
#ifdef LGCA
  LogPrintf("lattice gas enabled\n");
#endif
#ifdef CYCLAGE_HOR
  //LogPrintf("support for periodic boundary conditions\n");
#endif
#ifdef OPENMP
  LogPrintf("OpenMP enabled\n");
#endif
#ifdef USE_AVX
  LogPrintf("AVX enabled\n");
#endif
#ifdef PARALLEL_AUTOMATA
  //LogPrintf("\"pat\" multithreading supported\n");
#endif
#ifdef PARALLEL
  LogPrintf("parallel execution with MPI\n");
#endif
#ifdef __CYGWIN__
  LogPrintf("compiled under Cygwin\n");
#endif

  //LogPrintf("Copyright (C) 2011-%.4s\n", VER_DAT);

  if (argc == 1){
    usage_no_input();
    exit(-4);
  }

  flag_no_input = (*argv[1] == '-');

  general_options(argc, argv);

  int opt_h_all = opt_h || opt_hm || opt_ht;
  if (opt_h_all){
    rescal_usage();

    if (opt_h){
      show_general_options();
      show_view_options();
    }

    if (opt_hm){
      rescal_params();
      param_usage();
    }

    if(opt_ht){
      templates_usage();
    }

    printf("\nPlease see README file for instructions ");
    printf("(also see http://www.ipgp.fr/rescal).\n");
  }

  if (flag_no_input){
    if (!opt_h_all) usage_no_input();
    exit(0);
  }


  /// is execution deterministic ?
  int dflag = 1;
#ifdef DETERMINISTIC
#ifdef PARALLEL_AUTOMATA
  if (mode_pat) dflag = 0;
#endif // PARALLEL_AUTOMATA
#ifdef PARALLEL
  if (mode_par) dflag = 0;
#endif // PARALLEL
#else
#if defined(LGCA) && defined(OPENMP)
  dflag = 0;
#endif // LGCA
#endif // DETERMINISTIC
  if (dflag)
    LogPrintf("deterministic execution\n")
  else
    LogPrintf("non-deterministic execution\n");

  if (!opt_nv){
#if (GLIB_MAJOR_VERSION <= 2) && (GLIB_MINOR_VERSION<32)
    g_thread_init(NULL); /// deprecated function since glib 2.32
#endif
    gdk_threads_init();
    gtk_init (&argc, &argv);
  }
  /*else gtk_init_check (&argc, &argv);*/

  rescal_init(argv[1]);
  view_init(argc, argv);
  callbacks_init();

  //if (opt_nv) rescal(0);

#ifdef GUI
  if (!opt_nv) {
    create_window();
    gcallback_flash(0);
  }
#endif

  if (opt_log) {
    if (dlog_delay.unit==UNIT_COMP){
      dp_info = (DumpPar){DUMP_LOG, (int)dlog_delay.val};
      log_info();
      g_timeout_add_seconds((guint)(dlog_delay.val), gcallback_dump, &dp_info);
    }
    else if (dlog_delay.unit==UNIT_T0){
      schedul_add_reg_event(EVT_LOG, dlog_delay.val, 0);
    }
  }

  if (opt_dcsp && !flag_dcsp_auto) {
    if (dcsp_delay.unit==UNIT_COMP){
      dp_csp = (DumpPar){DUMP_CSP, (int)dcsp_delay.val};
      gcallback_dump(&dp_csp);
      g_timeout_add_seconds((guint)(dcsp_delay.val*60), gcallback_dump, &dp_csp);
    }
    else if (dcsp_delay.unit==UNIT_T0){
      schedul_add_reg_event(EVT_SAVE_CSP, dcsp_delay.val, 0);
    }
  }

  if (opt_dhpp && !flag_dhpp_auto /*&& !csphpp_flag*/) {
    if (dhpp_delay.unit==UNIT_T0){
      schedul_add_reg_event(EVT_SAVE_HPP, dhpp_delay.val, 0);
    }
  }

  if (opt_dbin) {
    //dump_terre(DUMP_BIN, dbin_delay);
    dp_bin = (DumpPar){DUMP_BIN, dbin_delay};
    gcallback_dump(&dp_bin);
    g_timeout_add_seconds(dbin_delay*60, gcallback_dump, &dp_bin);
  }

#if defined(USE_LIBPNG) || defined(USE_GD)
  view_dump_init();
#endif

  if (opt_dpng && !flag_dpng_auto) {
    if (dpng_delay.unit==UNIT_COMP){
      dump_image_inter((int)dpng_delay.val, "png");
      dp_png = (DumpPar){DUMP_PNG, (int)dpng_delay.val};
      g_timeout_add_seconds((guint)dpng_delay.val, gcallback_dump, &dp_png);
    }
    else if (dpng_delay.unit==UNIT_T0){
      schedul_add_reg_event(EVT_SAVE_PNG, dpng_delay.val, 0);
    }
  }
  if (opt_djpeg) {
    dump_image_inter(djpeg_delay, "jpeg");
    dp_jpeg = (DumpPar){DUMP_JPEG, djpeg_delay};
    g_timeout_add_seconds(djpeg_delay, gcallback_dump, &dp_jpeg);
  }

  if (opt_stop) {
    g_timeout_add_seconds_full(G_PRIORITY_DEFAULT, stop_delay*60+15, gcallback_stop, 0, NULL);
    //gtk_timeout_add(stop_delay*60000+15000, callback_stop, 0);
  }
  if (opt_quit){
    g_timeout_add_seconds_full(G_PRIORITY_DEFAULT, 1, gcallback_quit, 0, NULL);
  }

  int nRetVal;
  pthread_t pth;
  nRetVal = pthread_create( &pth, 0, rescal_thread, 0);

  /*
  GThread *pth;
  GError *err = NULL;
  g_thread_init(NULL);
  pth = g_thread_create((GThreadFunc)rescal_thread, 0, FALSE, &err);
  */

  //gtk_idle_add(idle, NULL);

  gtk_main ();

#ifdef LOG_FILE
  fclose(log_file);
#endif

  return 0;
}

void read_delay(char *str, DumpDelay *pd)
{
  int n;
  if (strstr(str, "t0")){
    pd->unit = UNIT_T0;
    if (sscanf(str, "%lft0", &pd->val)!=1) {ErrPrintf("ERROR: invalid delay value %s\n",str); exit(-1);}
  }
  else{
    pd->unit = UNIT_COMP;
    if (sscanf(str, "%d", &n)!=1) {ErrPrintf("ERROR: invalid delay value %s\n",str); exit(-1);}
    pd->val = n;
  }
  //LogPrintf("delay: %lf (%d)\n", pd->val, pd->unit);
}

void general_options(int argc, char *argv[])
{
  int i;
  DumpDelay dd;

  /// log the command line options
  if (!flag_no_input){
    LogPrintf("args: ");
    for(i=1; i<argc; i++) LogPrintf("%s ", argv[i]);
    LogPrintf("\n");
  }

  /// parse the command line options
  for(i=1; i<argc; i++){
    if (!strcmp(argv[i],"-h") || !strcmp(argv[i],"--help"))
      opt_h = 1;
    else if (!strcmp(argv[i],"-hm"))
      opt_hm = 1;
    else if (!strcmp(argv[i],"-ht"))
      opt_ht = 1;
    else if (!strcmp(argv[i],"-nv"))
      opt_nv = 1;
    else if (!strcmp(argv[i],"-fr")){
      int frame_rate = atoi(argv[++i]);
      frame_delay = (frame_rate<100) ? 1000/frame_rate : 10;
    }
    else if (!strcmp(argv[i],"-info")){
      opt_log = 1;
#ifdef PARALLEL
      if (proc_id > 0) opt_log = 0;
#endif
      dlog_delay.unit = UNIT_COMP;
      dlog_delay.val = 60;
    }
    else if (!strcmp(argv[i],"-dlog")){
      opt_log = 1;
#ifdef PARALLEL
      if (proc_id > 0) opt_log = 0;
#endif
      read_delay(argv[++i], &dlog_delay);
      //LogPrintf("log delay: %d\n", (int)dlog_delay.val);
    }
    else if (!strcmp(argv[i],"-dcsp") || !strcmp(argv[i],"-dcsphpp")){
      if (!strcmp(argv[i],"-dcsphpp")) csphpp_flag = 1;
      opt_dcsp = 1;
      i++;
      if (!strcmp(argv[i],"auto") || !strcmp(argv[i],"AUTO")){
        flag_save_auto = 1;
        flag_dcsp_auto = 1;
        LogPrintf("automatic saving of CSP files before each rotation\n");
      }
      else{
        read_delay(argv[i], &dcsp_delay);
      }
    }
    else if (!strcmp(argv[i],"-dhpp")){
      opt_dhpp = 1;
      i++;
      if (!strcmp(argv[i],"auto") || !strcmp(argv[i],"AUTO")){
        flag_save_auto = 1;
        flag_dhpp_auto = 1;
        LogPrintf("automatic saving of HPP files before each rotation\n");
      }
      else{
        read_delay(argv[i], &dhpp_delay);
      }
    }
    else if (!strcmp(argv[i],"-dbin")){
      opt_dbin = 1;
      dbin_delay = atoi(argv[++i]);
    }
    else if (!strcmp(argv[i],"-dpng")){
      opt_dpng = 1;
      i++;
      if (!strcmp(argv[i],"auto") || !strcmp(argv[i],"AUTO")){
        flag_save_auto = 1;
        flag_dpng_auto = 1;
        LogPrintf("automatic saving of PNG files before each rotation\n");
      }
      else{
        read_delay(argv[i], &dpng_delay);
      }
    }
#ifndef USE_LIBPNG
    else if (!strcmp(argv[i],"-djpeg")){
      opt_djpeg = 1;
      djpeg_delay = atoi(argv[++i]);
      if (!djpeg_delay) djpeg_delay = 1;
    }
#endif
    else if (!strcmp(argv[i],"-stop")){
      read_delay(argv[++i], &dd);
      if (dd.unit==UNIT_COMP){
        opt_stop = 1;
        stop_delay = dd.val; //atoi(argv[++i]);
      }
      else if (dd.unit==UNIT_T0){
        stop_delay_t0 = dd.val + MIN(dd.val/1e6,1); //0.001;
      }
      LogPrintf("stop delay: %s\n", argv[i]);
    }
    else if (!strcmp(argv[i],"-stoptime")){
      read_delay(argv[++i], &dd);
      if (dd.unit==UNIT_T0){
        stop_time = dd.val + MIN(dd.val/1e6,1); //0.001;
      }
      else if (dd.unit==UNIT_COMP){
        ErrPrintf("ERROR: bad unit in -stoptime option\n");
        exit(-1);
      }
      LogPrintf("stop time: %s\n", argv[i]);
    }
    else if (!strcmp(argv[i],"-q")){
      opt_quit = 1;
    }
    else if (!strcmp(argv[i],"-s") || !strcmp(argv[i],"-g")){
      graine = atoi(argv[++i]);
    }
#ifdef PARALLEL_AUTOMATA
    else if (!strcmp(argv[i],"-pat")){
      mode_pat = 1;
      LogPrintf("parallel automata with threads\n");
    }
#endif
#ifdef PARALLEL
    else if (!strcmp(argv[i],"-par")){
      mode_par = 1;
    }
#endif
    /*else if (*argv[i]=='-'){
      ErrPrintf("ERROR: unknown option %s\n", argv[i]);
      exit(-5);
    }*/
  }
  if (opt_dcsp && opt_dbin){
    ErrPrintf("ERROR: -dcsp and -dbin options are mutually exclusive\n");
    exit(-1);
  }
  //if (csphpp_flag && opt_dhpp) opt_dhpp = 0;
#ifndef GUI
  opt_nv = 1;
#endif
}

void show_general_options()
{
  printf("\nGENERAL OPTIONS\n");
  printf("  -h \t\t display usage information\n");
  printf("  -hm \t\t display parameters of the model\n");
  printf("  -ht \t\t display templates of the model\n");
#ifdef GUI
  printf("  -nv \t\t no window mode\n");
#endif
  //printf("  -info \t logging data every minute\n");
  printf("  -dlog <n> \t log data every <n> seconds\n");
  printf("  -dlog <f>t0 \t log data with delay in t0 unit (float value)\n");
  printf("  -dcsp <n> \t generate a CSP file every <n> minutes\n");
  printf("  -dcsp <f>t0 \t generate a CSP file with delay in t0 unit (float value)\n");
#ifdef ROTATIONS
  printf("  -dcsp auto \t generate a CSP file when a rotation occurs\n");
#endif
  printf("  -dhpp <f>t0 \t generate HPP and GRAD_VEL files with delay in t0 unit (float value)\n");
#ifdef ROTATIONS
  printf("  -dhpp auto \t generate HPP and GRAD_VEL files when a rotation occurs\n");
#endif
  printf("  -dcsphpp <f>t0 \t generate CSP, HPP and GRAD_VEL files with delay in t0 unit (float value)\n");
#ifdef ROTATIONS
  printf("  -dcsphpp auto \t generate CSP, HPP and GRAD_VEL files when a rotation occurs\n");
#endif
  printf("  -dbin <n> \t generate a binary file every <n> minutes\n");
  printf("  -dpng <n> \t generate a PNG image every <n> seconds\n");
  printf("  -dpng <f>t0\t generate PNG images with delay in t0 unit (float value)\n");
#ifdef ROTATIONS
  printf("  -dpng auto \t generate a PNG image when a rotation occurs\n");
#endif
#ifndef USE_LIBPNG
  printf("  -djpeg <n> \t generate a Jpeg image every <n> seconds\n");
#endif
  printf("  -stop <n> \t quit after <n> minutes\n");
  printf("  -stop <f>t0 \t quit after running <f> t0 in simulated time\n");
  printf("  -stoptime <f>t0 \t quit when simulated time reaches <f> t0\n");
  printf("  -q \t\t quit at the end of simulation\n");
#ifdef GUI
  printf("  -fr <n> \t frame rate (fps)\n");
#endif
  printf("  -s <n> \t random seed\n");
#ifdef PARALLEL_AUTOMATA
  printf("  -pat \t\t parallel computation of lattice-gas and stochastic automata with threading\n");
#endif
//#ifdef MODEL_DUN
//  printf("  -ro <n> : rotation d'angle aleatoire toutes les <n> minutes\n");
//#endif
#ifdef PARALLEL
  printf("  -par \t\t parallel mode enabled by domain decomposition using MPI\n");
  //printf("  -pars : mode parallele, process serveur\n");
  //printf("  -parc <i> : mode parallele, process client numero <i>\n");
#endif
  fflush(stdout);
}

void usage_no_input()
{
  printf("no input file\n");
  printf("Try '-h' option for usage\n");
}

//#ifdef GTK_OLD_SYNTAX
#if GTK2 && (GTK_MINOR_VERSION<14)
void g_timeout_add_seconds(int nsec, GtkFunction gcallback_dump, gpointer data)
{
  gtk_timeout_add(nsec*1000, gcallback_dump, data);
}

void g_timeout_add_seconds_full(int prior, int nsec, GtkFunction gcallback_dump, gpointer data, void *notify)
{
  gtk_timeout_add(nsec*1000, gcallback_dump, data);
}
#endif
