/* ReSCAL - Event scheduler whithin the simulation
 *
 * Copyright (C) 2021
 *
 * Author: Olivier Rozier <rozier@ipgp.fr>
 *
 * Code based on dissol program,
 * by Eduardo Sepulveda <edo@espci.fr>
 *
 * This file is part of ReSCAL.
 *
 * ReSCAL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <memory.h>
#include <math.h>
#include <time.h>

#include "defs.h"
#include "macros.h"
#include "simul.h"
#include "simul_models.h"
#include "schedul.h"

extern double csp_time;                  // time variable for the cellular space

Evt_sched sched_gen; /// scheduler for general events
Evt_sched sched_mod; /// scheduler for the events of the model
int total_nb_evt=0; /// total number of events in any scheduler

const int evt_fam_gen[EVT_GEN_NUMBER] =  EVT_FAM_GEN;
const char *evt_name_gen[EVT_GEN_NUMBER] =  EVT_NAME_GEN;
const char *evt_name_mod[EVT_TOTAL_NUMBER-EVT_GEN_NUMBER] =  EVT_NAME_MOD;
char *evt_names[EVT_TOTAL_NUMBER];

void simul_save_auto();
double get_next_time_threshold();


void schedul_init()
{
  LogPrintf("initialization of 'general' and 'model' schedulers\n");

  sched_gen.family = EF_GEN;
  sched_gen.name = "general";
  sched_gen.nb_events = 0;
  sched_gen.next_time_threshold = 0.0;
  //sched_gen.epsilon = 0.1;  /// time tolerance for the first occurence of an event (sometimes useful when continuing a previous simulation)

  sched_mod.family = EF_MODEL;
  sched_mod.name = "model";
  sched_mod.nb_events = 0;
  sched_mod.next_time_threshold = 0.0;
  //sched_mod.epsilon = 0.0;

  LogPrintf("number of event types: %d\n", EVT_TOTAL_NUMBER);

  int i;
  for(i=0; i<EVT_TOTAL_NUMBER; i++){
    evt_names[i] = (char *) ( (i < EVT_GEN_NUMBER) ? evt_name_gen[i] : evt_name_mod[i-EVT_GEN_NUMBER] );
    //LogPrintf("type %d: %s (%s)\n", i, evt_names[i], (i < EVT_GEN_NUMBER) ? sched_gen.name : sched_mod.name);
  }
}


/// Declare single event triggered at time: evt_time
void schedul_add_event(int type, double evt_time)
{
  Evt_sched *p_sched = NULL;
  Evt *p_evt = NULL;

  Assert(type < EVT_TOTAL_NUMBER);

  p_sched = (type<EVT_GEN_NUMBER) ? &sched_gen : &sched_mod;

  Assert(p_sched->nb_events < NB_EVT_MAX);

  /// initialize singular event
  p_evt = p_sched->evt + (p_sched->nb_events++);
  p_evt->family = (type<EVT_GEN_NUMBER) ? evt_fam_gen[type] : EF_MODEL;
  p_evt->type = type;
  p_evt->name = evt_names[type];
  //p_evt->regular = 0;
  p_evt->periodicity = EP_SINGLE;
  p_evt->active = 1;
  p_evt->delay = 0;
  p_evt->time_threshold = evt_time;
  if ((p_sched->nb_events == 1) || (p_sched->next_time_threshold > p_evt->time_threshold)){
    p_sched->next_time_threshold = p_evt->time_threshold;
  }

  total_nb_evt++;

  LogPrintf("unique event %s (%s) added at time %f (t0)\n", p_evt->name, p_sched->name, evt_time);
  //LogPrintf("number of events: %d\n", p_sched->nb_events);

  /// update time_threshold of scheduler
  Assert(p_evt->time_threshold >= csp_time);
  if ((p_sched->nb_events == 1) || (p_sched->next_time_threshold > p_evt->time_threshold)){
    p_sched->next_time_threshold = p_evt->time_threshold;
  }
}


/// Declare regular event with constant delay (in t0) between two consecutive events
/// First event triggered at csp_time if offset=0, otherwise a time threshold is set
/// to some multiple of delay above csp_time, according to offset.
void schedul_add_reg_event(int type, double delay, int offset)
{
  Evt_sched *p_sched = NULL;
  Evt *p_evt = NULL;

  //LogPrintf("schedul_add_reg_event: type=%d\n", type);

  Assert(type < EVT_TOTAL_NUMBER);
  Assert(delay > 0.0);

  p_sched = (type<EVT_GEN_NUMBER) ? &sched_gen : &sched_mod;

  Assert(p_sched->nb_events < NB_EVT_MAX);

  /// initialize regular event
  p_evt = p_sched->evt + (p_sched->nb_events++);
  p_evt->family = (type<EVT_GEN_NUMBER) ? evt_fam_gen[type] : EF_MODEL;
  p_evt->type = type;
  p_evt->name = evt_names[type];
  //p_evt->regular = 1;
  p_evt->periodicity = EP_REGULAR;
  p_evt->active = 1;
  p_evt->delay = delay;
  //p_evt->time_threshold = init_time + offset*delay; //delay*ceil((csp_time-p_sched->epsilon)/delay);
  p_evt->time_threshold = (offset == 0) ? csp_time : delay*(floor(csp_time/delay) + offset);
  //if (offset>0) LogPrintf("first event %d at time_threshold=%f, with csp_time=%f, delay=%f, offset=%d\n", type, p_evt->time_threshold, csp_time, delay, offset);

  /// update time_threshold of scheduler
  Assert(p_evt->time_threshold >= csp_time);
  if ((p_sched->nb_events == 1) || (p_sched->next_time_threshold > p_evt->time_threshold)){
    p_sched->next_time_threshold = p_evt->time_threshold;
  }

  total_nb_evt++;

  LogPrintf("regular event %s (%s) added with delay %f (t0)\n", p_evt->name, p_sched->name, delay);
  //LogPrintf("next time threshold: %f\n", p_sched->next_time_threshold);
  //LogPrintf("number of events: %d\n", p_sched->nb_events);
}


/// Declare non regular event starting at time t0
void schedul_add_non_reg_event(int type, double t0)
{
  Evt_sched *p_sched = NULL;
  Evt *p_evt = NULL;

  Assert(type < EVT_TOTAL_NUMBER);
  Assert(t0 >= csp_time);

  p_sched = (type<EVT_GEN_NUMBER) ? &sched_gen : &sched_mod;

  Assert(p_sched->nb_events < NB_EVT_MAX);

  /// initialize non-regular event
  p_evt = p_sched->evt + (p_sched->nb_events++);
  p_evt->family = (type<EVT_GEN_NUMBER) ? evt_fam_gen[type] : EF_MODEL;
  p_evt->type = type;
  p_evt->name = evt_names[type];
  p_evt->periodicity = EP_NON_REGULAR;
  p_evt->active = 1;
  p_evt->delay = 0;
  p_evt->time_threshold = t0;

  total_nb_evt++;

  LogPrintf("non regular event %s (%s) added starting at %f (t0)\n", p_evt->name, p_sched->name, t0);

  /// update time_threshold of scheduler
  Assert(p_evt->time_threshold >= csp_time);
  if ((p_sched->nb_events == 1) || (p_sched->next_time_threshold > p_evt->time_threshold)){
    p_sched->next_time_threshold = p_evt->time_threshold;
  }
}


void trigger_event(Evt *p_evt)
{
  double delay = 0; /// time delay until next non-regular event of the model

  //LogPrintf("event %s triggered (%d)\n", p_evt->name, p_evt->family); /// debug

  if (p_evt->family == EF_MODEL){
    /// handle a "model" event
    delay = simul_model_event(p_evt->type);
    //LogPrintf("trigger_event: delay=%f\n", delay);
  }
  else{
    /// handle a "general" event
    simul_general_event(p_evt->type);
  }

  if (p_evt->periodicity == EP_SINGLE){
    /// make event inactive
    p_evt->active = 0;
  }
  else if (p_evt->periodicity == EP_REGULAR){
    /// set next time threshold to a multiple of the delay
    p_evt->time_threshold = p_evt->delay*ceil(csp_time/p_evt->delay);
    while (p_evt->time_threshold <= csp_time) p_evt->time_threshold += p_evt->delay;
  }
  else if (p_evt->periodicity == EP_NON_REGULAR){
    /// update time threshold
    //LogPrintf("set delay %f for non-regular event %s\n", delay, p_evt->name);
    Assert(delay>0);
    p_evt->time_threshold += delay;
  }
}

int is_scheduled(int type)
{
  Evt_sched *p_sched = NULL;
  Evt *p_evt;
  int i;

  p_sched = (type<EVT_GEN_NUMBER) ? &sched_gen : &sched_mod;

  p_evt = p_sched->evt;
  for(i=0; i<p_sched->nb_events; i++, p_evt++){
    if ((p_evt->type == type) && p_evt->active) return(1);
  }

  return(0);
}

/// Check family of events
void schedul_chk_family(int evt_fam)
{
  int i;
  Evt *p_evt;
  Evt_sched *p_sched = (evt_fam == EF_MODEL) ? &sched_mod : &sched_gen;

  //LogPrintf("simul_sched_family: family %s, nb_events = %d, next_time_threshold = %f, csp_time = %f\n", p_sched->name, p_sched->nb_events, p_sched->next_time_threshold, csp_time); /// debug

  if ((p_sched->nb_events) && (csp_time >= p_sched->next_time_threshold)){
    //LogPrintf("time threshold reached: %f \n", p_sched->next_time_threshold);
    p_evt = p_sched->evt;
    for(i=0; i<p_sched->nb_events; i++, p_evt++){
      //LogPrintf("check evt %d: %.10f t0\n", i, p_sched->evt[i].time_threshold);
      if (p_evt->active && (p_evt->family == evt_fam) && (p_sched->next_time_threshold == p_evt->time_threshold)){

        /// trigger event
        //LogPrintf("trigger evt %d (i=%d)\n", p_evt->type, i);
        trigger_event(p_evt);
      }
    }
    /// update next time threshold
    p_evt = p_sched->evt;
    p_sched->next_time_threshold = p_evt->time_threshold;
    for(i=1, p_evt++; i<p_sched->nb_events; i++, p_evt++){
      if (p_evt->active && (p_sched->next_time_threshold > p_evt->time_threshold))
        p_sched->next_time_threshold = p_evt->time_threshold;
    }
    //LogPrintf("next time threshold: %f\n", p_sched->next_time_threshold);
  }
}


/// Check all scheduled events
void schedul_chk_all()
{
  schedul_chk_family(EF_GEN);
  schedul_chk_family(EF_MODEL);
  schedul_chk_family(EF_GEN_CA);
}


/// Return the next time threshold from all schedulers
double get_next_time_threshold()
{
  if (sched_gen.nb_events && sched_mod.nb_events)
    return Min(sched_gen.next_time_threshold, sched_mod.next_time_threshold);
  else if (sched_gen.nb_events)
    return sched_gen.next_time_threshold;
  else if (sched_mod.nb_events)
    return sched_mod.next_time_threshold;
  else
    return csp_time+1e10;
}


/*inline int get_nb_events(int fam)
{
  if (fam == EF_GEN) return sched_gen.nb_events;
  if (fam == EF_MODEL) return sched_mod.nb_events;
  if (fam == EF_ALL) return total_nb_evt;
  ErrPrintf("ERROR: get_nb_events (%d)\n", fam);
  exit(-2);
}*/


inline int get_total_nb_events()
{
  return total_nb_evt;
}

